/*
 *
 */
package org.vamsas.client.simpleclient;

import java.util.Hashtable;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.vamsas.client.IClient;
import org.vamsas.client.IClientAppdata;
import org.vamsas.client.IClientDocument;
import org.vamsas.client.VorbaId;
import org.vamsas.client.Vobject;
import org.vamsas.objects.core.AppData;
import org.vamsas.objects.core.ApplicationData;
import org.vamsas.objects.core.User;
import org.vamsas.objects.core.VAMSAS;
import org.vamsas.objects.core.VamsasDocument;
import org.vamsas.objects.utils.AppDataReference;

/**
 * Maintains a collection of vamsas objects, appdatas and states, and provides api for a SimpleClient's client.
 * @author jimp 
 */
public class ClientDocument extends org.vamsas.client.ClientDocument implements IClientDocument {
  private static Log log = LogFactory.getLog(ClientDocument.class);
  private VamsasDocument doc;
  protected SimpleClient sclient;
  protected VamsasArchive archive = null;
  /**
   * indicate if new data has been incorporated
   */
  private boolean isModified = false;
  /**
   * Public method for internal use by SimpleClient.
   * @return true if document has been modified.
   */
  public boolean isModified() {
    return isModified;
  }
  /**
   *
   *  prepare Application-side dataset from the vamsas Document archive
   * @param doc - the dataset
   * @param docHandler - the sessionFile IO handler
   * @param Factory - the source of current and new vorbaIds
   * @param sclient - the simpleclient instance
   */
  protected ClientDocument(VamsasDocument doc, VamsasArchive docHandler, IdFactory Factory, SimpleClient sclient) {
    super(Factory.getVorbaIdHash(), Factory);
    
    /**
     * prepare Application-side dataset from the vamsas Document archive
     */
    this.sclient = sclient;
    archive = docHandler;
    this.doc = doc;
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see org.vamsas.client.IClientDocument#getObject(org.vamsas.client.VorbaId)
   */
  public Vobject getObject(VorbaId id) {
    // TODO: look up id in document Vobject
    // retrieve Vobject and return
    return null;
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see org.vamsas.client.IClientDocument#getObjects(org.vamsas.client.VorbaId[])
   */
  public Vobject[] getObjects(VorbaId[] ids) {
    // TODO: getObject in bulk
    return null;
  }
  /**
   * internal reference to single copy of Document Roots array
   */
  private VAMSAS[] _VamsasRoots=null;
  /*
   * (non-Javadoc)
   * LATER: currently there is only one Vector of roots ever passed to client - decide if this is correct (means this is not thread safe and may behave unexpectedly)
   * @see org.vamsas.client.IClientDocument#getVamsasRoots()
   */
  public VAMSAS[] getVamsasRoots() {
    if (_VamsasRoots!=null)
      return _VamsasRoots;
    VAMSAS[] roots = doc.getVAMSAS();
    if (roots == null) {
      // Make a new one to return to client : TODO: Decide if this is correct
      _VamsasRoots = new VAMSAS[] { new VAMSAS() };
    } else {
      _VamsasRoots = new VAMSAS[roots.length];
      for (int r=0;r<roots.length; r++)
        _VamsasRoots[r] = roots[r];
    }
    return _VamsasRoots;
  }
  
  private int _contains(VAMSAS root, VAMSAS[] docRoots) {
    if (root==null)
      return -1;
    if (docRoots==null || docRoots.length==0)
      return -1;
    String r_id = root.getId();
    for (int i=0,j=docRoots.length; i<j; i++)
      if (docRoots[i]==root || (docRoots[i]!=null && docRoots[i].getId().equals(r_id)))
        return i;
    return -1;
  }
/**
 * verify that newr version is really an intact version of the 
 * @param newVersion (may be modified)
 * @param oldVersion 
 * @return true if newVersion is a valid root that preserves original references
 */
  private boolean isValidUpdate(VAMSAS newVersion, final VAMSAS oldVersion) {
    // LATER: IMPLEMENT
    // ideal - this cascades down the two structures, ensuring that all ID'd objects in one are present in the other.
    if (oldVersion==newVersion) {
      // may be a virgin root element.
      if (!newVersion.isRegistered())
        _registerObject(newVersion);
      // Should retrieve original version and compare - unless local hashes can be used to determine if resultSet has been truncated.
      // just do internal validation for moment.
      if (newVersion.isValid())
        return true;
      return false;
    } else {
      // redundant ? if (oldVersion.is__stored_in_document())
      if (!newVersion.isRegistered())
        _registerObject(newVersion);
      if (newVersion.isValid())
        return true;
    }
    return false;
  }
  /**
   * merge old and new root vectors
   * @param newr This array may be written to
   * @param original
   * @param the client document (usually this) which this root set belongs to.
   * @return merged vector of vamsas roots
   */
  private VAMSAS[] _combineRoots(VAMSAS[] newr, final VAMSAS[] original, ClientDocument modflag) {
    Vector rts = new Vector();
    boolean modified=false;
    for (int i=0,j=original.length; i<j; i++) {
      int k = _contains(original[i], newr);
      if (k>-1) {
        if (isValidUpdate(newr[k], original[i])) {
          modified=true;
          rts.add(newr[k]);
          newr[k]=null;
        } else {
          // LATER: try harder to merge ducument roots.
          log.warn("Couldn't merge new VAMSAS root "+newr[k].getId());
          newr[k] = null; // LATER: this means we ignore mangled roots. NOT GOOD
        }
      } else {
        // add in order.
        rts.add(original[i]);
      }
    }
    // add remaining (new) roots
    for (int i=0,j=newr.length; i<j; i++) {
      if (newr[i]!=null) {
        rts.add(newr[i]);
        modified=true;
      }
    }
    newr = new VAMSAS[rts.size()];
    for (int i=0,j=rts.size(); i<j; i++)
      newr[i] = (VAMSAS) rts.get(i);
    if (modflag!=null)
      modflag.isModified = modified;
    return newr;
  }
  
  /**
   * update the document with new roots.
   * LATER: decide: this affects the next call to getVamsasRoots()
   */
  public void setVamsasRoots(VAMSAS[] newroots) {
    
    // are we dealing with same array ?
    if (_VamsasRoots!=newroots) {
      // merge roots into 
    }
    // extract root objects
    if (newroots != null) {
      // check newroots for objects that were present in the old document
      // check to see if the 'old' objects have been modified
      // if they have ? we overwrite them with their new version, ensuring that
      // provenance is updated.
      // if they haven't ? do nothing ?
      
      for (int i = 0, k = newroots.length; i < k; i++) {
        if (newroots[i].isRegistered()) {
          // easy - just check if anything has changed and do provenance
          Vobject oldversion = getObject(newroots[i].getVorbaId());
          if (oldversion instanceof VAMSAS) {
            // LATER: appropriate merging behaviour when two clients have improperly modified the same Vobject independently.
            if (newroots[i].get__last_hash() != newroots[i].hashCode()) {
              // client has modified this Vobject since last retrieval.
              if (newroots[i].get__last_hash() != oldversion.get__last_hash()) {
                // Vobject has been modified by another client since this
                // client's
                // last access to document.
              }
            }
          } else {
            throw new Error(
                "SimpleClient error when using setVamsasRoots : The vorbaId for Vobject "
                + i
                + " does not refer to an Vobject of type VAMSAS in the current document!");
          }
        } else {
          if (!newroots[i].is__stored_in_document()) {
            // check if Vobject is modified
            if (newroots[i].get__last_hash() != newroots[i].hashCode()) {
              // it is - so we add newroots[i] as a new Vobject, with updated
              // provenance.
            } else {
              // do nothing
              newroots[i] = null;
            }
          } else {
            // just add newroots[i] as a new Vobject in the document
            // - with appropriate provenance.
          }
        }
      }
    }
  }
  
  /* (non-Javadoc)
   * LATER: decide: this affects the next call to getVamsasRoots()
   * @see org.vamsas.client.IClientDocument#addVamsasRoot(org.vamsas.objects.core.VAMSAS)
   */
  public void addVamsasRoot(VAMSAS newroot) {
    VAMSAS[] newroots = _combineRoots(new VAMSAS[] {newroot}, _VamsasRoots, this);
    _VamsasRoots = newroots;  
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see org.vamsas.client.IClientDocument#registerObjects(org.vamsas.client.Vobject[])
   */
  public VorbaId[] registerObjects(Vobject[] unregistered) {
    if (unregistered!=null) {
      VorbaId ids[] = new VorbaId[unregistered.length];
      for (int i=0,k=unregistered.length; i<k; i++)
        if (unregistered[i]!=null) {
          log.warn("Null Vobject passed to registerObject[] at position "+i);
          return null;
        } else {
          ids[i]=registerObject(unregistered[i]);
        }
      return ids;
    }
    return null;
  }
  /* (non-Javadoc)
   * @see org.vamsas.client.IClientDocument#registerObject(org.vamsas.client.Vobject)
   */
  public VorbaId registerObject(Vobject unregistered) {
    if (unregistered!=null)
      return _registerObject(unregistered);
    log.warn("Null Vobject passed to registerObject.");
    return null;
  }
  /**
   * IClientAppdata instance - if it exists.
   */
  SimpleClientAppdata scappd = null;
  /* (non-Javadoc)
   * @see org.vamsas.client.IClientDocument#getClientAppdata()
   */
  public IClientAppdata getClientAppdata() {
    if (scappd==null) {
      log.debug("Creating new SimpleClientAppdata instance for "+sclient.getSessionHandle());
      scappd = new SimpleClientAppdata(this);
      if (scappd==null) {
        // LATER: may not need this as a warning message.
        log.warn("Null appdata object for "+sclient.getSessionHandle());
      } else {
        log.debug("Created SimpleClientAppdata successfully.");
      }
    } else {
      log.debug("Returning existing SimpleClientAppdata reference.");
    }
    return scappd;
  }
  /* (non-Javadoc)
   * @see java.lang.Object#finalize()
   */
  protected void finalize() throws Throwable {
    log.debug("Garbage collecting on ClientDocument instance.");
    if (scappd!=null) {
      scappd.finalize();
      scappd = null;
    }
    
    super.finalize();
  }
  /**
   * access the vamsas document
   * @return the session's vamsas document
   */
  protected VamsasDocument getVamsasDocument() {
    return doc;
  }
  /**
   * returns the read-only IO interface for the vamsas document Jar file
   * @return
   */
  protected VamsasArchiveReader getVamsasArchiveReader() {
    try {
      return archive.getOriginalArchiveReader();
    } catch (Exception e) {
      log.warn("Unable to create OriginalArchiveReader!", e);
    }
    return null;
  }
}
